# -*- coding: utf-8 -*-

require_relative 'spec_helper'
require 'pathname'

$VERBOSE = nil

include General

module RobustExcelOle

describe Workbook do

  before(:all) do
    excel = Excel.new(:reuse => true)
    open_books = excel == nil ? 0 : excel.Workbooks.Count
    puts "*** open books *** : #{open_books}" if open_books > 0
    Excel.kill_all
  end

  before do
    @dir = create_tmpdir
    @simple_file = @dir + '/workbook.xls'
    @pathname_file = Pathname(@dir) + 'workbook.xls'
    @simple_save_file = @dir + '/workbook_save.xls'
    @different_file = @dir + '/different_workbook.xls'
    @simple_file_other_path = @dir + '/more_data/workbook.xls'
    @another_simple_file = @dir + '/another_workbook.xls'
    @linked_file = @dir + '/workbook_linked.xlsm'
    @simple_file_xlsm = @dir + '/workbook.xls'
    @simple_file_xlsx = @dir + '/workbook.xlsx'
    @simple_file1 = @simple_file
    @simple_file_other_path1 = @simple_file_other_path
    @simple_save_file1 = @simple_save_file
  end

  after do
    Excel.kill_all
    rm_tmp(@dir)
  end

  describe "create file" do
    context "with standard" do
      it "open an existing file" do
        expect {
          @book = Workbook.open(@simple_file)
        }.to_not raise_error
        @book.should be_a Workbook
        @book.close
      end
    end

    context "with pathname" do
      it "open an existing file" do
        expect {
          @book = Workbook.open(@pathname_file)
        }.to_not raise_error
        @book.should be_a Workbook
        @book.close
      end
    end
  end

  describe "Book" do

    context "with standard options" do
      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should say that it lives" do
        @book.should be_alive
      end
    end

  end

  describe "open and new" do

    context "with standard" do

      it "should yield identical objects" do
        book1 = Workbook.open(@simple_file1)
        ole_workbook1 = book1.ole_workbook
        book2 = Workbook.new(ole_workbook1)
        book1.equal?(book2).should be true
      end

    end

  end

  describe "connecting to unknown workbooks" do

    context "with one unknown workbook" do

      before do
        ole_e1 = WIN32OLE.new('Excel.Application')
        ws = ole_e1.Workbooks
        abs_filename = General.absolute_path(@simple_file1)
        @ole_wb = ws.Open(abs_filename)
      end

      it "should connect to an unknown workbook" do
        Workbook.open(@simple_file1) do |book|
          book.filename.should == @simple_file1
          book.should be_alive
          book.should be_a Workbook
          book.excel.ole_excel.Hwnd.should == @ole_wb.Application.Hwnd
          Excel.instance_count.should == 1
        end
      end
    end
  end

  describe "Workbook::save" do

    it "should save a file, if it is open" do
      @book = Workbook.open(@simple_file)
      @sheet = @book.sheet(1)
      @book.add_sheet(@sheet, :as => 'a_name')
      @new_sheet_count = @book.ole_workbook.Worksheets.Count
      expect {
        Workbook.save(@simple_file)
      }.to_not raise_error
      @book.ole_workbook.Worksheets.Count.should ==  @new_sheet_count
      @book.close
    end

    it "should not save a file, if it is not open" do
      @book = Workbook.open(@simple_file)
      @sheet = @book.sheet(1)
      @book.add_sheet(@sheet, :as => 'a_name')
      @new_sheet_count = @book.ole_workbook.Worksheets.Count
      @book.close(:if_unsaved => :forget)
      expect {
        Workbook.save(@simple_file)
      }.to_not raise_error
    end

  end

  describe "Workbook::save_as" do
    
    it "should save to 'simple_save_file.xls'" do
      book = Workbook.open(@simple_file1)
      Workbook.save_as(@simple_file1, @simple_save_file1, :if_exists => :overwrite)
      File.exist?(@simple_save_file1).should be true
    end
  end

  describe "Workbook::close" do

    it "should close the book if it is open" do
      book = Workbook.open(@simple_file1)
      Workbook.close(@simple_file1)
      book.should_not be_alive
    end

    it "should not close the book if it is not open" do
      book = Workbook.open(@simple_file1, :visible => true)
      book.close
      Workbook.close(@simple_file1)
      book.should_not be_alive
    end

    it "should raise error if the book is unsaved and open" do
      book = Workbook.open(@simple_file1)
      sheet = book.sheet(1)
      book.add_sheet(sheet, :as => 'a_name')
      expect{
        Workbook.close(@simple_file1)
      }.to raise_error(WorkbookNotSaved, /workbook is unsaved: "workbook.xls"/)
      expect{
        Workbook.close(@simple_file, :if_unsaved => :raise)
      }.to raise_error(WorkbookNotSaved, /workbook is unsaved: "workbook.xls"/)
    end

    it "should save and close the book" do
      book = Workbook.open(@simple_file1)
      sheet_count = book.ole_workbook.Worksheets.Count
      sheet = book.sheet(1)
      book.add_sheet(sheet, :as => 'a_name')
      ole_workbook = book.ole_workbook
      excel = book.excel
      excel.Workbooks.Count.should == 1
      Workbook.close(@simple_file1, {:if_unsaved => :save})
      excel.Workbooks.Count.should == 0
      book.ole_workbook.should == nil
      book.should_not be_alive
      #expect{ole_workbook.Name}.to raise_error(WIN32OLERuntimeError)
      new_book = Workbook.open(@simple_file1)
      begin
        new_book.ole_workbook.Worksheets.Count.should == sheet_count + 1
      ensure
        new_book.close
      end
    end
  end

  describe "open" do

    context "with various file formats" do

      it "should open linked workbook" do
        book = Workbook.open(@linked_file, :visible => true)
        book.close
      end
    end


    context "with class identifier 'Workbook'" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close rescue nil
      end

      it "should open in a new Excel" do
        book2 = Workbook.open(@simple_file, :force => {:excel => :new})
        book2.should be_alive
        book2.should be_a Workbook
        book2.excel.should_not == @book.excel 
        book2.should_not == @book
        @book.Readonly.should be false
        book2.Readonly.should be true
        book2.close
      end
    end

    context "lift a workbook to a Workbook object" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should yield an identical Workbook and set visible and displayalerts values" do
        workbook = @book.ole_workbook
        new_book = Workbook.new(workbook, :visible => true)
        new_book.excel.displayalerts = true
        new_book.should be_a Workbook
        new_book.should be_alive
        new_book.should == @book
        new_book.filename.should == @book.filename
        new_book.excel.should == @book.excel
        new_book.should === @book
        new_book.excel.Visible.should be true
        new_book.excel.DisplayAlerts.should be true
        new_book.close
      end

    end

    context "with standard options" do
      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should say that it lives" do
        @book.should be_alive
      end
    end

    context "with identity transperence" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should yield identical Workbook objects for identical Excel books when reopening" do
        @book.should be_alive
        @book.close
        @book.should_not be_alive
        book2 = Workbook.open(@simple_file)
        book2.should === @book
        book2.should be_alive
        book2.close
      end

      it "should yield different Workbook objects when reopening in a new Excel" do
        @book.should be_alive
        old_excel = @book.excel
        @book.close
        @book.should_not be_alive
        book2 = Workbook.open(@simple_file, :force => {:excel => :new})
        book2.should_not === @book
        book2.should be_alive
        book2.excel.should_not == old_excel
        book2.close
      end
    end

    context "with :force_excel" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close rescue nil
      end

      it "should open in a new Excel" do
        book2 = Workbook.open(@simple_file, :force => {:excel => :new})
        book2.should be_alive
        book2.should be_a Workbook
        book2.excel.should_not == @book.excel 
        book2.should_not == @book
        @book.Readonly.should be false
        book2.Readonly.should be true
        book2.close
      end

      it "should open in a given Excel, provide identity transparency, because book can be readonly, such that the old and the new book are readonly" do
        book2 = Workbook.open(@simple_file1, :force => {:excel => :new})
        book2.excel.should_not == @book.excel
        book3 = Workbook.open(@simple_file1, :force => {:excel => :new})
        book3.excel.should_not == book2.excel
        book3.excel.should_not == @book.excel
        book2.close
        book3.close
        @book.close
        book4 = Workbook.open(@simple_file1, :force => {:excel => book2.excel}, :read_only => true)
        book4.should be_alive
        book4.should be_a Workbook
        book4.excel.should == book2.excel
        book4.ReadOnly.should be true
        book4.should == book2
        book4.close
        book5 = Workbook.open(@simple_file1, :force => {:excel => book2}, :read_only => true)
        book5.should be_alive
        book5.should be_a Workbook
        book5.excel.should == book2.excel
        book5.ReadOnly.should be true
        book5.should == book2
        book5.close
        book3.close
      end
    end

    context "with :default_excel" do

      before do
        excel = Excel.new(:reuse => false)
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close rescue nil
      end

      it "should reopen the book in the excel instance where it was opened before" do
        excel = Excel.new(:reuse => false)
        @book.close
        book2 = Workbook.open(@simple_file)
        book2.should be_alive
        book2.should be_a Workbook
        book2.excel.should == @book.excel
        book2.excel.should_not == excel
        book2.filename.should == @book.filename
        @book.should be_alive
        book2.should == @book
        book2.close
      end

      it "should open a new excel, if the book cannot be reopened" do
        @book.close
        new_excel = Excel.new(:reuse => false)
        book2 = Workbook.open(@different_file, :default => {:excel => :new})
        book2.should be_alive
        book2.should be_a Workbook
        book2.excel.should_not == new_excel
        book2.excel.should_not == @book.excel
        book2.close
      end

      it "should open in a given Excel provided as Excel, Workbook, or WIN32OLE representing an Excel or Workbook" do
        book2 = Workbook.open(@another_simple_file)
        different_file1 = @different_file
        book3 = Workbook.open(different_file1, :default => {:excel => book2.excel})
        book3.excel.should === book2.excel
        book3.close
        book4 = Workbook.open(different_file1, :default => {:excel => book2}) 
        book4.excel.should === book2.excel
        book4.close
        book5 = Workbook.open(different_file1, :default_excel => book2.ole_workbook)
        book5.excel.should ===  book2.excel
        book5.close
        win32ole_excel1 = WIN32OLE.connect(book2.ole_workbook.Fullname).Application
        book6 = Workbook.open(different_file1, :default => {:excel => win32ole_excel1})
        book6.excel.should === book2.excel
        book6.close
      end


    end

    context "with :if_unsaved" do

      before do
        @book = Workbook.open(@simple_file)
        @sheet = @book.sheet(1)
        @book.add_sheet(@sheet, :as => 'a_name')
      end

      after do
        @book.close(:if_unsaved => :forget)
        @new_book.close rescue nil
      end

      it "should raise an error, if :if_unsaved is :raise" do
        expect {
          @new_book = Workbook.open(@simple_file, :if_unsaved => :raise)
        }.to raise_error(WorkbookNotSaved, /workbook is already open but not saved: "workbook.xls"/)
      end

      it "should let the book open, if :if_unsaved is :accept" do
        expect {
          @new_book = Workbook.open(@simple_file, :if_unsaved => :accept)
          }.to_not raise_error
        @book.should be_alive
        @new_book.should be_alive
        @new_book.should == @book
      end

      context "with :if_unsaved => :alert or :if_unsaved => :excel" do
        before do
         @key_sender = IO.popen  'ruby "' + File.join(File.dirname(__FILE__), '/helpers/key_sender.rb') + '" "Microsoft Excel" '  , "w"
        end

        after do
          @key_sender.close
        end

        it "should open the new book and close the unsaved book, if user answers 'yes'" do
          # "Yes" is the  default. --> language independent
          @key_sender.puts "{enter}"
          @new_book = Workbook.open(@simple_file1, :if_unsaved => :alert)
          @new_book.should be_alive
          @new_book.filename.downcase.should == @simple_file1.downcase
          @book.should_not be_alive
        end

        it "should not open the new book and not close the unsaved book, if user answers 'no'" do
          # "No" is right to "Yes" (the  default). --> language independent
          # strangely, in the "no" case, the question will sometimes be repeated three times
          #@book.excel.Visible = true
          @key_sender.puts "{right}{enter}"
          @key_sender.puts "{right}{enter}"
          @key_sender.puts "{right}{enter}"
          #expect{
            Workbook.open(@simple_file, :if_unsaved => :alert)    
          #  }.to raise_error(UnexpectedREOError)
          @book.should be_alive
          @book.Saved.should be false
        end

        it "should open the new book and close the unsaved book, if user answers 'yes'" do
          # "Yes" is the  default. --> language independent
          @key_sender.puts "{enter}"
          @new_book = Workbook.open(@simple_file1, :if_unsaved => :excel)
          @new_book.should be_alive
          @new_book.filename.downcase.should == @simple_file1.downcase
          #@book.should_not be_alive
        end

        it "should not open the new book and not close the unsaved book, if user answers 'no'" do
          # "No" is right to "Yes" (the  default). --> language independent
          # strangely, in the "no" case, the question will sometimes be repeated three times
          #@book.excel.Visible = true
          @key_sender.puts "{right}{enter}"
          @key_sender.puts "{right}{enter}"
          @key_sender.puts "{right}{enter}"
          #expect{
            Workbook.open(@simple_file, :if_unsaved => :excel)
          #}.to raise_error(UnexpectedREOError)
          @book.should be_alive
          @book.Saved.should be false
        end

      end
    end

    context "with :if_blocked" do

      for i in 1..2 do

        context "with and without reopen" do

          before do   
            if i == 1 then 
              book_before = Workbook.open(@simple_file)
              book_before.close
            end
            @book = Workbook.open(@simple_file_other_path1)
            @sheet_count = @book.ole_workbook.Worksheets.Count
            @sheet = @book.sheet(1)
            @book.add_sheet(@sheet, :as => 'a_name')
          end

          after do
            @book.close(:if_unsaved => :forget)
            @new_book.close rescue nil
          end

          it "should save the old book, close it, and open the new book, if :if_blocked is :save" do
            @new_book = Workbook.open(@simple_file1, :if_blocked => :save)
            @book.should_not be_alive
            @new_book.should be_alive
            @new_book.filename.downcase.should == @simple_file1.downcase
            old_book = Workbook.open(@simple_file_other_path1, :if_blocked => :forget)
            old_book.ole_workbook.Worksheets.Count.should ==  @sheet_count + 1
            old_book.close
          end

          it "should raise an error, if the old book is unsaved, and close the old book and open the new book, 
              if :if_blocked is :close_if_saved" do
            expect{
              @new_book = Workbook.open(@simple_file1, :if_blocked => :close_if_saved)
            }.to raise_error(WorkbookBlocked, /workbook with the same name in a different path is unsaved/)
            @book.save
            @new_book = Workbook.open(@simple_file1, :if_blocked => :close_if_saved)
            @book.should_not be_alive
            @new_book.should be_alive
            @new_book.filename.downcase.should == @simple_file1.downcase
            old_book = Workbook.open(@simple_file_other_path1, :if_blocked => :forget)
            old_book.ole_workbook.Worksheets.Count.should ==  @sheet_count + 1
            old_book.close
          end
        end
      end
    end

    context "with :if_obstructed" do

      for i in 1..2 do

        context "with and without reopen" do

          before do   
            if i == 1 then 
              book_before = Workbook.open(@simple_file)
              book_before.close
            end
            @book = Workbook.open(@simple_file_other_path1)
            @sheet_count = @book.ole_workbook.Worksheets.Count
            @sheet = @book.sheet(1)
            @book.add_sheet(@sheet, :as => 'a_name')
          end

          after do
            @book.close(:if_unsaved => :forget)
            @new_book.close rescue nil
          end

          it "should save the old book, close it, and open the new book, if :if_obstructed is :save" do
            @new_book = Workbook.open(@simple_file1, :if_obstructed => :save)
            @book.should_not be_alive
            @new_book.should be_alive
            @new_book.filename.downcase.should == @simple_file1.downcase
            old_book = Workbook.open(@simple_file_other_path1, :if_obstructed => :forget)
            old_book.ole_workbook.Worksheets.Count.should ==  @sheet_count + 1
            old_book.close
          end

          it "should raise an error, if the old book is unsaved, and close the old book and open the new book, 
              if :if_obstructed is :close_if_saved" do
            expect{
              @new_book = Workbook.open(@simple_file1, :if_obstructed => :close_if_saved)
            }.to raise_error(WorkbookBlocked, /workbook with the same name in a different path is unsaved/)
            @book.save
            @new_book = Workbook.open(@simple_file1, :if_obstructed => :close_if_saved)
            @book.should_not be_alive
            @new_book.should be_alive
            @new_book.filename.downcase.should == @simple_file1.downcase
            old_book = Workbook.open(@simple_file_other_path1, :if_obstructed => :forget)
            old_book.ole_workbook.Worksheets.Count.should ==  @sheet_count + 1
            old_book.close
          end
        end
      end
    end

    context "with non-existing file" do

      it "should create a workbook" do
        File.delete @simple_save_file rescue nil
        book = Workbook.create(@simple_save_file)
        book.should be_a Workbook
        book.ReadOnly.should be false
        book.close
        File.exist?(@simple_save_file).should be true
      end

      it "should create a workbook" do
        File.delete @simple_save_file rescue nil
        book = Workbook.open(@simple_save_file, :if_absent => :create)
        book.should be_a Workbook
        book.ReadOnly.should be false
        book.close
        File.exist?(@simple_save_file).should be true
      end

      it "should raise an exception by default" do
        File.delete @simple_save_file rescue nil
        expect {
          Workbook.open(@simple_save_file)
        }.to raise_error(FileNotFound, /not found/ )
        #  "\nHint: If you want to create a new file, use option :if_absent => :create or Workbook::create")
      end
    end

    context "with :read_only" do
      
      it "should reopen the book with writable (unsaved changes from readonly will not be saved)" do
        book = Workbook.open(@simple_file1, :read_only => true)
        book.ReadOnly.should be true
        book.should be_alive
        sheet = book.sheet(1)
        old_cell_value = sheet[1,1]
        sheet[1,1] = sheet[1,1] == "foo" ? "bar" : "foo"
        book.Saved.should be false
        new_book = Workbook.open(@simple_file1, :read_only => false, :if_unsaved => :forget)
        new_book.ReadOnly.should be false 
        new_book.should be_alive
        book.should_not be_alive   
        new_sheet = new_book.sheet(1)
        new_cell_value = new_sheet[1,1]
        new_cell_value.should == old_cell_value
      end


    context "with block" do
      it 'block parameter should be instance of Workbook' do
        Workbook.open(@simple_file) do |book|
          book.should be_a Workbook
        end
      end
    end
  end

  describe "reopen" do

    context "with standard" do
      
      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should reopen the closed book" do
        @book.should be_alive
        book1 = @book
        @book.close
        @book.should_not be_alive
        @book.reopen
        @book.should be_a Workbook
        @book.should be_alive
        @book.should === book1
      end
    end
  end

  describe "type-lifting" do

    context "with standard" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should type-lift a win32ole workbook object to a workbook object with an open workbook" do
        workbook = @book.ole_workbook
        book1 = Workbook.new(workbook)
        book1.should be_a Workbook
        book1.should be_alive
        book1.should == @book
      end
    end
  end

  describe "visible" do

    it "should preserve :visible if they are not set" do
      excel1 = Excel.create(:visible => true)
      book1 = Workbook.open(@simple_file)
      book1.excel.Visible.should be true
      book1.close
    end

    it "should preserve :visible if they are not set" do
      excel1 = Excel.create
      book1 = Workbook.open(@simple_file, :visible => true)
      book1.excel.Visible.should be true
    end

    it "should preserve :visible if they are not set" do
      excel1 = Excel.create(:visible => true)
      book1 = Workbook.open(@simple_file, :default => {:excel => :new})
      book1.excel.should_not == excel1
      book1.excel.Visible.should be false
    end

    it "should preserve :visible if they are not set" do
      excel1 = Excel.create(:visible => true)
      excel2 = Excel.create(:visible => true)
      book1 = Workbook.open(@different_file, :force => {:excel => excel2})
      book1.excel.Visible.should be true
      book1.close
    end

    it "should let an open Workbook open" do
      @book = Workbook.open(@simple_file, :visible => true)
      Workbook.unobtrusively(@simple_file) do |book|
        book.should be_a Workbook
        book.should be_alive
        book.excel.should == @book.excel
        book.excel.Visible.should be true
      end        
      @book.should be_alive
      @book.should be_a Workbook
      @book.excel.Visible.should be true
      @book.close(:if_unsaved => :forget)
      @book2.close(:if_unsaved => :forget) rescue nil
    end
  end


  describe "unobtrusively" do

    # @private
    def unobtrusively_ok? 
      Workbook.unobtrusively(@simple_file) do |book|
        book.should be_a Workbook
        sheet = book.sheet(1)
        sheet[1,1] = sheet[1,1] == "foo" ? "bar" : "foo"
        book.should be_alive
        book.Saved.should be false
      end
    end

    context "with no open book" do

      it "should open unobtrusively in a new Excel" do
        expect{ unobtrusively_ok? }.to_not raise_error
      end
    end

    context "with two running excel instances" do
      before :all do
        Excel.kill_all
      end

      before do
        @excel1 = Excel.new(:reuse => false)
        @excel2 = Excel.new(:reuse => false)
      end

      after do
        #Excel.close_all
        begin
          @excel1.close
          @excel2.close 
        rescue ExcelREOError => msg
          # puts "ExcelREOError: #{msg.message}" if msg.message =~ /Excel instance not alive or damaged/
        end
      end

      it "should open unobtrusively in a new Excel" do
        Workbook.unobtrusively(@simple_file, :if_closed => :current) do |book|
          book.should be_a Workbook
          book.should be_alive
          book.excel.should == @excel1
          book.excel.should_not == @excel2
        end
      end

      it "should open unobtrusively in a given Excel" do
        Workbook.unobtrusively(@simple_file, :if_closed => @excel2) do |book|
          book.should be_a Workbook
          book.should be_alive
          book.excel.should_not == @excel1
          book.excel.should     == @excel2
        end
      end
    end
  
    context "with an open book" do

      before do
        @book = Workbook.open(@simple_file1)
      end

      after do
        @book.close(:if_unsaved => :forget)
        @book2.close(:if_unsaved => :forget) rescue nil
      end

      it "should let an open Workbook open if two workbooks have been opened and one has been closed and opened again" do
        book2 = Workbook.open(@different_file, :force => {:excel => :new})
        @book.close
        book2.close
        @book.reopen
        Workbook.unobtrusively(@simple_file1) do |book|
          book.should be_a Workbook
          book.should be_alive
          book.excel.should == @book.excel
        end        
        @book.should be_alive
        @book.should be_a Workbook
      end

      it "should let a saved book saved" do
        @book.Saved.should be true
        @book.should be_alive
        sheet = @book.sheet(1)
        old_cell_value = sheet[1,1]
        unobtrusively_ok?
        @book.Saved.should be true
        @book.should be_alive
        sheet = @book.sheet(1)
        sheet[1,1].should_not == old_cell_value
      end

     it "should let the unsaved book unsaved" do
        sheet = @book.sheet(1)
        sheet[1,1] = sheet[1,1] == "foo" ? "bar" : "foo"
        old_cell_value = sheet[1,1]
        @book.Saved.should be false
        unobtrusively_ok?
        @book.should be_alive
        @book.Saved.should be false
        @book.close(:if_unsaved => :forget)
        @book2 = Workbook.open(@simple_file1)
        sheet2 = @book2.sheet(1)
        sheet2[1,1].should_not == old_cell_value
      end
    end
    
    context "with a closed book" do
      
      before do
        @book = Workbook.open(@simple_file1)
      end

      after do
        @book.close(:if_unsaved => :forget)
      end

      it "should let the closed book closed by default" do
        sheet = @book.sheet(1)
        old_cell_value = sheet[1,1]
        @book.close
        @book.should_not be_alive
        unobtrusively_ok?
        @book.should_not be_alive
        book2 = Workbook.open(@simple_file1)
        sheet = book2.sheet(1)
        sheet[1,1].should_not == old_cell_value
      end

      # The bold reanimation of the @book
      it "should use the excel of the book and keep open the book" do
        excel = Excel.new(:reuse => false)
        sheet = @book.sheet(1)
        old_cell_value = sheet[1,1]
        @book.close
        @book.should_not be_alive
        Workbook.unobtrusively(@simple_file1, :keep_open => true) do |book|
          book.should be_a Workbook
          book.excel.should == @book.excel
          book.excel.should_not == excel
          sheet = book.sheet(1)
          cell = sheet.range([1,1])
          sheet[1,1] = cell.Value == "foo" ? "bar" : "foo"
          book.Saved.should be false
        end
        @book.should be_alive
        @book.close
        new_book = Workbook.open(@simple_file1)
        sheet = new_book.sheet(1)
        sheet[1,1].should_not == old_cell_value
      end

      it "should use the excel of the book and keep open the book" do
        excel = Excel.new(:reuse => false)
        sheet = @book.sheet(1)
        old_cell_value = sheet[1,1]
        @book.close
        @book.should_not be_alive
        Workbook.unobtrusively(@simple_file1, :if_closed => :current) do |book|
          book.should be_a Workbook
          book.should be_alive
          book.excel.should == @book.excel
          book.excel.should_not == excel
          sheet = book.sheet(1)
          cell = sheet.range([1,1])
          sheet[1,1] = cell.Value == "foo" ? "bar" : "foo"
          book.Saved.should be false
        end
        @book.should_not be_alive
        new_book = Workbook.open(@simple_file1)
        sheet = new_book.sheet(1)
        sheet[1,1].should_not == old_cell_value
      end
    end

    
    context "with a read_only book" do

      before do
        @book = Workbook.open(@simple_file1, :read_only => true)
      end

      after do
        @book.close
      end
    end

    context "with a virgin Workbook class" do
      before do
        # @private
        class Workbook 
          @@bookstore = nil
        end
      end
      it "should work" do
        expect{ unobtrusively_ok? }.to_not raise_error
      end
    end

    context "with a book never opened before" do
      before do
        # @private
        class Workbook   
          @@bookstore = nil
        end
        other_book = Workbook.open(@different_file)
      end
      it "should open the book" do
        expect{ unobtrusively_ok? }.to_not raise_error
      end
    end

    context "with block result" do
      before do
        @book1 = Workbook.open(@simple_file)
      end

      after do
        @book1.close(:if_unsaved => :forget)
      end      

      it "should yield the block result true" do
        result = 
          Workbook.unobtrusively(@simple_file) do |book| 
            @book1.Saved.should be true
          end
        result.should == true
      end
    end

    context "with several Excel instances" do

      before do
        @book1 = Workbook.open(@simple_file1)
        @book2 = Workbook.open(@simple_file1, :force_excel => :new)
        @book1.Readonly.should == false
        @book2.Readonly.should == true
        old_sheet = @book1.sheet(1)
        @old_cell_value = old_sheet[1,1]
        @book1.close
        @book2.close
        @book1.should_not be_alive
        @book2.should_not be_alive
      end

      it "should open unobtrusively the closed book in the most recent Excel where it was open before" do      
        Workbook.unobtrusively(@simple_file1, :if_closed => :current) do |book| 
          book.excel.should == @book2.excel
          book.excel.should_not == @book1.excel
          book.ReadOnly.should == false
          sheet = book.sheet(1)
          cell = sheet.range([1,1])
          sheet[1,1] = cell.Value == "foo" ? "bar" : "foo"
          book.Saved.should be false
        end
        new_book = Workbook.open(@simple_file1)
        sheet = new_book.sheet(1)
        sheet[1,1].should_not == @old_cell_value
      end

    end
  end
    
  describe "for_reading, for_modifying" do

    context "open unobtrusively for reading and modifying" do

      before do
        @book = Workbook.open(@simple_file1)
        sheet = @book.sheet(1)
        @old_cell_value = sheet[1,1]
        @book.close
      end

      it "should not change the value" do
        Workbook.for_reading(@simple_file) do |book|
          book.should be_a Workbook
          book.should be_alive
          book.Saved.should be true  
          sheet = book.sheet(1)
          cell = sheet.range([1,1])
          sheet[1,1] = cell.Value == "foo" ? "bar" : "foo"
          book.Saved.should be false
          book.excel.should == @book.excel
        end
        new_book = Workbook.open(@simple_file1, :visible => true)
        sheet = new_book.sheet(1)
        sheet[1,1].should == @old_cell_value
      end


      it "should not change the value and use the hidden Excel instance" do
        new_excel = Excel.new(:reuse => false)
        Workbook.for_reading(@simple_file1, :if_closed => :new) do |book|
          sheet = book.sheet(1)
          cell = sheet.range([1,1])
          sheet[1,1] = cell.Value == "foo" ? "bar" : "foo"
          book.excel.should_not == @book.excel
          book.excel.should_not == new_excel
          book.excel.properties[:visible].should be false
          book.excel.properties[:displayalerts].should == :if_visible
        end
        new_book = Workbook.open(@simple_file1, :visible => true)
        sheet = new_book.sheet(1)
        sheet[1,1].should == @old_cell_value
      end

      it "should change the value" do
        Workbook.for_modifying(@simple_file1) do |book|
          sheet = book.sheet(1)
          cell = sheet.range([1,1])
          sheet[1,1] = cell.Value == "foo" ? "bar" : "foo"
          book.excel.should == @book.excel
        end
        new_book = Workbook.open(@simple_file1, :visible => true)
        sheet = new_book.sheet(1)
        sheet[1,1].should_not == @old_cell_value
      end
    end
  end

  describe "send methods to workbook" do

    context "with standard" do
      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should send Saved to workbook" do
        @book.Saved.should be true
      end
    end
  end

  describe "nameval, set_nameval, [], []=" do
  
    before do
      @book1 = Workbook.open(@another_simple_file)
    end

    after do
      @book1.close(:if_unsaved => :forget)
    end   

    it "should return value of a range" do
      @book1.namevalue_global("new").should == "foo"
      @book1.namevalue_global("one").should == 1
      @book1.namevalue_global("firstrow").should == [[1,2]]        
      @book1.namevalue_global("four").should == [[1,2],[3,4]]
      @book1.namevalue_global("firstrow").should_not == "12"
      @book1.namevalue_global("firstcell").should == "foo"        
    end

    it "should return value of a range via []" do
      @book1["new"].should == "foo"
      @book1["one"].should == 1
      @book1["firstrow"] == [[1,2]]        
      @book1["four"].should == [[1,2],[3,4]]
      @book1["firstrow"].should_not == "12"
      @book1["firstcell"].should == "foo"        
    end

    it "should set value of a range" do
      @book1.set_namevalue_global("new", "bar")
      @book1.namevalue_global("new").should == "bar"
    end

    it "should set value of a range via []=" do
      @book1["new"] = "bar"
      @book1.namevalue_global("new").should == "bar"
    end

    #it "should evaluate a formula" do
    #  @book1.namevalue_global("named_formula").should == 4      
    #end

    #it "should evaluate a formula via []" do
    #  @book1["named_formula"].should == 4      
    #end

    #it "should return default value if name not defined" do
    #  @book1.namevalue_global("foo", :default => 2).should == 2
    #end

  end

  describe "close" do

    context "with unsaved read_only book" do
      before do
        @book = Workbook.open(@simple_file, :read_only => true)
        @sheet_count = @book.ole_workbook.Worksheets.Count
        @book.add_sheet(@sheet, :as => 'a_name')
      end

      it "should close the unsaved book without error and without saving" do
        expect{
          @book.close
          }.to_not raise_error
        new_book = Workbook.open(@simple_file)
        new_book.ole_workbook.Worksheets.Count.should ==  @sheet_count
        new_book.close
      end
    end

    context "with unsaved book" do
      before do
        @book = Workbook.open(@simple_file)
        @sheet_count = @book.ole_workbook.Worksheets.Count
        @book.add_sheet(@sheet, :as => 'a_name')
        @sheet = @book.sheet(1)
      end

      after do
        @book.close(:if_unsaved => :forget) rescue nil
      end

      it "should raise error by default" do
        expect{
          @book.close(:if_unsaved => :raise)
        }.to raise_error(WorkbookNotSaved, /workbook is unsaved: "workbook.xls"/)
      end

      it "should save the book before close with option :save" do
        ole_workbook = @book.ole_workbook
        excel = @book.excel
        excel.Workbooks.Count.should == 1
        @book.close(:if_unsaved => :save)
        excel.Workbooks.Count.should == 0
        @book.ole_workbook.should == nil
        @book.should_not be_alive
        #expect{
         # ole_workbook.Name}.to raise_error(WIN32OLERuntimeError)
        new_book = Workbook.open(@simple_file)
        begin
          new_book.ole_workbook.Worksheets.Count.should == @sheet_count + 1
        ensure
          new_book.close
        end
      end
    end
  end

  describe "save" do

    context "with simple save" do
      
      it "should save for a file opened without :read_only" do
        @book = Workbook.open(@simple_file)
        @book.add_sheet(@sheet, :as => 'a_name')
        @new_sheet_count = @book.ole_workbook.Worksheets.Count
        expect {
          @book.save
        }.to_not raise_error
        @book.ole_workbook.Worksheets.Count.should ==  @new_sheet_count
        @book.close
      end
    end

    context "with argument" do
      before do
        Workbook.open(@simple_file) do |book|
          book.save_as(@simple_save_file, :if_exists => :overwrite)
        end
      end

      it "should save to 'simple_save_file.xlsx'" do
        File.exist?(@simple_save_file).should be true
      end
    end
  end

  describe "alive?, filename, ==, focus, saved" do

    context "with alive?" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should return true, if book is alive" do
        @book.should be_alive
      end

      it "should return false, if book is dead" do
        @book.close
        @book.should_not be_alive
      end

    end

    context "with filename" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
      end

      it "should return full file name" do
        @book.filename.should == @simple_file
      end

    end

    context "with ==" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close
        @new_book.close rescue nil
      end

      it "should be true with two identical books" do
        @new_book = Workbook.open(@simple_file)
        @new_book.should == @book
      end
    end

    context "with focus" do

      before do
        @key_sender = IO.popen  'ruby "' + File.join(File.dirname(__FILE__), '/helpers/key_sender.rb') + '" "Microsoft Excel" '  , "w"        
        @book = Workbook.open(@simple_file, :visible => true)
        @book2 = Workbook.open(@another_simple_file, :visible => true)
      end

      after do
        @book.close(:if_unsaved => :forget)
        @book2.close(:if_unsaved => :forget)
        @key_sender.close
      end
    end
  end

=begin
      it "should bring a book to focus" do
        sheet = @book.sheet(2)
        sheet.Activate
        sheet[2,3].Activate
        sheet2 = @book2.sheet(2)
        sheet2.Activate
        sheet2[3,2].Activate
        @book2.focus
        @key_sender.puts "{a}{enter}"
        sleep 0.2
        sheet2[3,2].Value.should == "a"
        @book.focus
        @book.Windows(1).Visible.should be true
        @book.Windows(@book.Name).Visible.should be true
        @key_sender.puts "{a}{enter}"
        sleep 0.2
        sheet[2,3].Value.should == "a"
        Excel.current.should == @book.excel
      end
    end
  end
=end

  describe "#add_sheet" do
    before do
      @book = Workbook.open(@simple_file)
      @sheet = @book.sheet(1)
    end

    after do
      @book.close(:if_unsaved => :forget)
    end

    context "only first argument" do
      it "should add worksheet" do
        @book.ole_workbook.Worksheets.Count.should == 3
        @book.add_sheet @sheet
        @book.ole_workbook.Worksheets.Count.should == 4
        #expect { @book.add_sheet @sheet }.to change{ @book.workbook.Worksheets.Count }.from(3).to(4)
      end

      it "should return copyed sheet" do
        sheet = @book.add_sheet @sheet
        copyed_sheet = @book.ole_workbook.Worksheets.Item(@book.ole_workbook.Worksheets.Count)
        sheet.name.should eq copyed_sheet.name
      end
    end

    context "with first argument" do

      context "with second argument is {:before => @book.sheet(3), :after => @sheet}" do
        it "should arguments in the first is given priority" do
          @book.add_sheet(@sheet, :before => @book.sheet(3), :after => @sheet)
          @book.Worksheets.Count.should == 4
        end
      end
    end

    context "without first argument" do
      context "second argument is {:as => 'new sheet'}" do
        it "should return new sheet" do
          @book.add_sheet(:as => 'new sheet').name.should eq 'new sheet'
        end
      end

      context "second argument is {:before => @sheet}" do
        it "should add the first sheet" do
          @book.add_sheet(:before => @sheet).name.should eq @book.sheet(1).name
        end
      end

      context "second argument is {:after => @sheet}" do
        it "should add the second sheet" do
          @book.add_sheet(:after => @sheet).name.should eq @book.sheet(2).name
        end
      end
    end

    context "without argument" do

      it "should return copyed sheet" do
        sheet = @book.add_sheet
        copyed_sheet = @book.ole_workbook.Worksheets.Item(@book.ole_workbook.Worksheets.Count)
        sheet.name.should eq copyed_sheet.name
      end
    end
  end

  describe "#names" do

    before do
      @book1 = Workbook.open(@another_simple_file)
    end

    it "should yield defined names" do
      @book1.names.should == ["another", "Sheet1!another_formula", "empty", "Sheet1!firstcell", 
                              "firstrow", "formula", "four", "Sheet1!localname", "Sheet2!localname", 
                              "named_formula", "new", "one", "Sheet1!simple"]
    end
      
  end

  describe "each" do

    before do
      @book = Workbook.open(@simple_file)
    end

    it "should do each" do
      @book.each do |sheet|
        sheet.should be_kind_of Worksheet
      end
    end

    it "should map" do
      @book.map{|s| s}.should == [@book.sheet(1), @book.sheet(2), @book.sheet(3)]
    end

    it "should concatenate" do
      names = ""
      @book.each.with_index{|s,i| names << "#{s.name} #{i} " }.to_a
      names.should == "Sheet1 0 Sheet2 1 Sheet3 2 "
    end

  end

  describe 'access sheet' do

    before do
      @book = Workbook.open(@simple_file)
    end

    after do
      @book.close
    end

    context "standard" do

      it 'with sheet name' do
        @book.sheet('Sheet1').should be_kind_of Worksheet
      end

      it 'with integer' do
        @book.sheet(1).should be_kind_of Worksheet
      end

      it 'with block' do
        @book.each do |sheet|
          sheet.should be_kind_of Worksheet
        end
      end

      it 'with each_with_index' do
        @book.each_with_index do |sheet,i|
          sheet.should be_kind_of Worksheet
        end
      end
    end

    describe "with retain_saved" do

      before do
        @book = Workbook.open(@simple_file)
      end

      after do
        @book.close(:if_unsaved => :forget)
      end

      it "should keep the save state 'unsaved'" do
        sheet = @book.sheet(1)
        sheet[1,1] = sheet[1,1] == "foo" ? "bar" : "foo"
        @book.Saved.should be false
        @book.retain_saved do
          sheet = @book.sheet(1)
          a = sheet[1,1]
          b = @book.visible
        end
        @book.Saved.should be false
      end

      it "should keep the save state 'unsaved' even when the workbook was saved before" do
        sheet = @book.sheet(1)
        sheet[1,1] = sheet[1,1] == "foo" ? "bar" : "foo"
        @book.Saved.should be false
        @book.retain_saved do
          @book.save
          @book.Saved.should be true
        end
        @book.Saved.should be false
      end
    end

=begin    
    context "with test what happens with save-status when setting calculation status" do

      it "should keep the save status" do
        book1 = Workbook.open(@simple_file, :visible => true)
        book1.Saved.should be true
        book2 = Workbook.open(@another_simple_file, :visible => true)
        book1.Saved.should be true
        book2.Saved.should be true
        sheet2 = book2.sheet(1)
        sheet2[1,1] = sheet2[1,1].Value == "foo" ? "bar" : "foo"
        book1.Saved.should be true
        book2.Saved.should be false
        book3 = Workbook.open(@different_file, :visible => true)
        book1.Saved.should be true
        book2.Saved.should be false
        book3.Saved.should be true
      end
    end
=end

    context 'open with block' do
      it {
        Workbook.open(@simple_file) do |book|
          book.sheet('Sheet1').should be_a Worksheet
        end
      }
    end
  end
end
end
end
